from datetime import date, timedelta

import django
from django.contrib.sessions.models import Session
from django.core.management import call_command
from django.db import connection
from django.test import TestCase
from django.test import override_settings
from django.utils import timezone

from django_scrubber import scrubbers
from django_scrubber.models import FakeData
from .models import DataFactory, DataToBeScrubbed


class TestScrubbers(TestCase):
    def test_empty_scrubber(self):
        data = DataFactory.create(first_name='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'first_name': scrubbers.Empty}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertEqual(data.first_name, '')

    def test_null_scrubber(self):
        data = DataFactory.create(last_name='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'last_name': scrubbers.Null}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertEqual(data.last_name, None)

    def test_hash_scrubber_max_length(self):
        data = DataFactory.create(first_name='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'first_name': scrubbers.Hash}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertNotEqual(data.first_name, 'Foo')
        self.assertEqual(
            len(data.first_name),
            DataToBeScrubbed._meta.get_field('first_name').max_length,
            'len(%s) != %d' % (data.first_name, DataToBeScrubbed._meta.get_field('first_name').max_length)
        )

    def test_hash_scrubber_textfield(self):
        data = DataFactory.create(description='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'description': scrubbers.Hash}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertNotEqual(data.description, 'Foo')

    def test_lorem_scrubber(self):
        data = DataFactory.create(description='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'description': scrubbers.Lorem}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertNotEqual(data.description, 'Foo')
        self.assertEqual(data.description[:11], 'Lorem ipsum')

    def test_faker_scrubber_charfield(self):
        data = DataFactory.create(last_name='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'last_name': scrubbers.Faker('last_name')}):
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertNotEqual(data.last_name, 'Foo')
        self.assertNotEqual(data.last_name, '')

    def test_faker_scrubber_with_provider_arguments(self):
        """
        Use this as an example for Faker scrubbers with parameters passed along
        """
        data = DataFactory.create(ean8='8')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'ean8': scrubbers.Faker('ean', length=8)}):
            call_command('scrub_data')
        data.refresh_from_db()

        # The EAN Faker will by default emit ean13, so this would fail if the parameter was ignored
        self.assertEquals(8, len(data.ean8))

        # Add a new scrubber for ean13
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'ean8': scrubbers.Faker('ean', length=13)}):
            call_command('scrub_data')
        data.refresh_from_db()

        # make sure it doesn't reuse the ean with length=8 scrubber
        self.assertEquals(13, len(data.ean8))

    def test_faker_scrubber_datefield(self):
        """
        Use this as an example for Scrubber's capability of optimistically Casting to the current field's type
        There is a bug with django < 2.1 and sqlite, that's why we don't run the test there.
        """
        if django.VERSION >= (2, 1) or connection.vendor != "sqlite":
            data = DataFactory.create(date_past=date.today())
            with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={
                    'date_past': scrubbers.Faker('past_date', start_date="-30d", tzinfo=None)}):
                call_command('scrub_data')
            data.refresh_from_db()

            self.assertGreater(date.today(), data.date_past)
            self.assertLess(date.today() - timedelta(days=31), data.date_past)

    def test_faker_scrubber_run_twice(self):
        """
        Use this as an example of what happens when you want to run the same Faker scrubbers twice
        """
        data = DataFactory.create(company='Foo')
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={
                'company': scrubbers.Faker('company')}):
            call_command('scrub_data')
            call_command('scrub_data')
        data.refresh_from_db()

        self.assertNotEqual(data.company, 'Foo')
        self.assertNotEqual(data.company, '')

    @override_settings(DEBUG=True)
    def test_faker_scrubber_run_clear_session_by_default(self):
        """
        Ensures that the session table will be emptied by default
        """
        # Create session object
        Session.objects.create(session_key='foo', session_data='Lorem ipsum', expire_date=timezone.now())

        # Sanity check
        self.assertTrue(Session.objects.all().exists())

        # Call command
        call_command('scrub_data')

        # Assertion that session table is empty now
        self.assertFalse(Session.objects.all().exists())

    @override_settings(DEBUG=True)
    def test_faker_scrubber_run_disable_session_clearing(self):
        """
        Ensures that the session table will be emptied by default
        """
        # Create session object
        Session.objects.create(session_key='foo', session_data='Lorem ipsum', expire_date=timezone.now())

        # Sanity check
        self.assertTrue(Session.objects.all().exists())

        # Call command
        call_command('scrub_data', keep_sessions=True)

        # Assertion that session table is empty now
        self.assertTrue(Session.objects.all().exists())

    @override_settings(DEBUG=True)
    def test_faker_scrubber_run_clear_faker_data_not_by_default(self):
        """
        Ensures that the session table will be emptied by default
        """
        # Create faker data object
        FakeData.objects.create(provider='company', content='Foo', provider_offset=1)

        # Sanity check
        self.assertTrue(FakeData.objects.filter(provider='company', content='Foo').exists())

        # Call command
        call_command('scrub_data')

        # Assertion that faker data still exists
        self.assertTrue(FakeData.objects.filter(provider='company', content='Foo').exists())

    @override_settings(DEBUG=True)
    def test_faker_scrubber_run_clear_faker_data_works(self):
        """
        Ensures that the session table will be emptied by default
        """
        # Create faker data object
        FakeData.objects.create(provider='company', content='Foo', provider_offset=1)

        # Sanity check
        self.assertTrue(FakeData.objects.filter(provider='company', content='Foo').exists())

        # Call command
        call_command('scrub_data', remove_fake_data=True)

        # Assertion that faker data still exists
        self.assertFalse(FakeData.objects.filter(provider='company', content='Foo').exists())



from django_scrubber.services.validator import ScrubberValidatorService

try:
    from unittest import mock
except ImportError:
    import mock

from django.core.management import call_command
from django.test import TestCase


class TestScrubValidator(TestCase):
    def test_scrub_validator_regular(self):
        with self.assertRaises(SystemExit):
            call_command('scrub_validation', verbosity=3)

    @mock.patch.object(ScrubberValidatorService, 'process')
    def test_scrub_validator_service_called(self, mocked_method):
        call_command('scrub_validation', verbosity=3)

        mocked_method.assert_called_once()



import sys

from django.core.management.base import BaseCommand

from django_scrubber.services.validator import ScrubberValidatorService


class Command(BaseCommand):

    def handle(self, *args, **options):
        service = ScrubberValidatorService()
        non_scrubbed_field_list = service.process()

        found_models = 0
        found_fields = 0
        if len(non_scrubbed_field_list):
            for model_path, affected_field_list in non_scrubbed_field_list.items():
                print(f'Model {model_path!r}:')
                found_models += 1
                for field in affected_field_list:
                    print(f"- {field}")
                    found_fields += 1

            print(f'{found_models} model(s) having {found_fields} unscrubbed field(s) detected.')
            sys.exit(1)

        print('No unscrubbed fields detected. Yeah!')



try:
    from unittest.mock import patch
except ImportError:
    from mock import patch
from io import StringIO

from django.core.management import call_command
from django.test import TestCase, override_settings

from django_scrubber import scrubbers
from django_scrubber.management.commands.scrub_data import _parse_scrubber_class_from_string, _get_model_scrubbers
from django.contrib.auth import get_user_model

User = get_user_model()


class TestScrubData(TestCase):
    def setUp(self):
        self.user = User.objects.create(first_name='test_first_name')

    def test_scrub_data(self):
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'first_name': scrubbers.Faker('first_name')}):
            call_command('scrub_data', verbosity=3)
        self.user.refresh_from_db()

        self.assertNotEqual(self.user.first_name, 'test_first_name')

    def test_scrub_data_debug_is_false(self):
        err = StringIO()

        with self.settings(DEBUG=False):
            call_command('scrub_data', stderr=err)
        output = err.getvalue()
        self.user.refresh_from_db()

        self.assertIn("This command should only be run with DEBUG=True, to avoid running on live systems", output)
        self.assertEqual(self.user.first_name, 'test_first_name')

    @override_settings(SCRUBBERS_STRICT_MODE=True)
    def test_scrub_data_strict_mode_enabled_scrubbing_blocked(self):
        err = StringIO()

        with self.settings(DEBUG=False):
            call_command('scrub_data', stderr=err)
        output = err.getvalue()
        self.user.refresh_from_db()

        self.assertIn("This command should only be run with DEBUG=True, to avoid running on live systems", output)
        self.assertEqual(self.user.first_name, 'test_first_name')

    def test_hash_simple_global_scrubber(self):
        with self.settings(DEBUG=True, SCRUBBER_GLOBAL_SCRUBBERS={'first_name': scrubbers.Hash}):
            call_command('scrub_data')
        self.user.refresh_from_db()

        self.assertNotEqual(self.user.first_name, 'test_first_name')

    def test_hash_simple_class_scrubber(self):
        class Scrubbers:
            first_name = scrubbers.Hash

        with self.settings(DEBUG=True), patch.object(User, 'Scrubbers', Scrubbers, create=True):
            call_command('scrub_data')
        self.user.refresh_from_db()

        self.assertNotEqual(self.user.first_name, 'test_first_name')

    def test_scrub_invalid_field(self):
        class Scrubbers:
            this_does_not_exist_382784 = scrubbers.Null

        with self.settings(DEBUG=True), patch.object(User, 'Scrubbers', Scrubbers, create=True):
            with self.assertWarnsRegex(
                Warning, 'Scrubber defined for User.this_does_not_exist_382784 but field does not exist'
            ):
                call_command('scrub_data')

    @override_settings(SCRUBBER_MAPPING={"auth.User": "tests.scrubbers.UserScrubbers"})
    def test_get_model_scrubbers_mapper_from_settings_used(self):
        with patch('django_scrubber.management.commands.scrub_data._parse_scrubber_class_from_string',
                   return_value={}) as mocked_method:
            with patch('django_scrubber.management.commands.scrub_data._get_fields', return_value=[]):
                test_scrubbers = _get_model_scrubbers(User)
        mocked_method.assert_called_once()
        self.assertEqual(test_scrubbers, {})

    def test_parse_scrubber_class_from_string_regular(self):
        class_type = _parse_scrubber_class_from_string('tests.test_models.TestDjangoScrubber')
        self.assertIsInstance(class_type, type)

    def test_parse_scrubber_class_from_string_wrong_path(self):
        with self.assertRaises(ImportError):
            _parse_scrubber_class_from_string('not.valid.path')

    def test_parse_scrubber_class_from_string_path_no_separator(self):
        with self.assertRaises(ImportError):
            _parse_scrubber_class_from_string('broken_path')



from django.apps import apps

from django_scrubber import settings_with_fallback


class ScrubberValidatorService:
    """
    Service to validate if all text-based fields are being scrubbed within your project and dependencies.
    """

    def process(self) -> dict:
        from django_scrubber.management.commands.scrub_data import _get_model_scrubbers

        scrubber_required_field_types = settings_with_fallback('SCRUBBER_REQUIRED_FIELD_TYPES')
        model_whitelist = settings_with_fallback('SCRUBBER_REQUIRED_FIELD_MODEL_WHITELIST')

        # Get a list of all registered models in your Django application
        model_list = apps.get_models()

        # Create a dictionary to store the fields of each model
        non_scrubbed_field_list = {}

        # Iterate over each model in the list
        for model in model_list:

            # Check if model is whitelisted
            if model._meta.label in model_whitelist:
                continue

            text_based_fields = []
            # Get the model's name and fields
            fields = model._meta.get_fields()

            # Gather list of all text-based files of the given model
            for field in fields:
                if type(field) in scrubber_required_field_types:
                    text_based_fields.append(field.name)

            # Get scrubber class
            scrubber_class = _get_model_scrubbers(model)

            # If we did find a scrubber class...
            if scrubber_class:
                # We check for every text-based field, if it's set to be scrubbed
                for scrubbed_field in scrubber_class:
                    if scrubbed_field.name in text_based_fields:
                        text_based_fields.remove(scrubbed_field.name)

            # Store per model all non-scrubbed, text-based fields
            if len(text_based_fields):
                non_scrubbed_field_list[model._meta.label] = text_based_fields

        return non_scrubbed_field_list



from builtins import str as text

import importlib
import logging

import faker

from django.db import router, connections
from django.db.models import Field, Func, Subquery, OuterRef
from django.db.models.functions import Concat as DjangoConcat, Cast
from django.db.utils import IntegrityError
from django.utils.translation import to_locale, get_language

from . import ScrubberInitError, settings_with_fallback

logger = logging.getLogger(__name__)


class FieldFunc(Func):
    """
    Base class for creating Func-like scrubbers.
    Unlike Func, may receive a Field object as first argument, in which case it populates self.extra with its __dict__.
    This enable derived classes to use the Field's attributes, either in methods or as interpolation variables in
    self.template.
    """
    def __init__(self, field, *args, **kwargs):
        if isinstance(field, Field):
            super(FieldFunc, self).__init__(field.name, *args, **kwargs)
            self.extra.update(field.__dict__)
            self.connection_setup(connections[router.db_for_write(field.model)])
        else:
            super(FieldFunc, self).__init__(field, *args, **kwargs)

    def connection_setup(self, db_connection):
        """
        This function is called when initializing the scrubber, and allows doing setup necessary to support certain DB
        vendors. It should be implemented by derived classes of FieldFunc.
        """
        pass


class Empty(FieldFunc):
    template = "''"


class Null(FieldFunc):
    template = 'NULL'


class Keep(FieldFunc):
    template = None


class Hash(FieldFunc):
    """
    Simple md5 hashing of content.
    If initialized with a Field object, will use its max_length attribute to truncate the generated hash.
    Otherwise, if initialized with a field name as string, will use the full hash length.
    """

    template = 'NULL'  # replaced during __init__
    arity = 1

    def __init__(self, *args, **kwargs):
        super(Hash, self).__init__(*args, **kwargs)
        if self.extra.get('max_length') is not None:
            self.template = 'SUBSTR(MD5(%(expressions)s), 1, %(max_length)s)'
        else:
            self.template = 'MD5(%(expressions)s)'

    def connection_setup(self, db_connection):
        if db_connection.vendor == 'sqlite':
            # add MD5 support for sqlite; this calls back to python and will probably have a performance impact
            import hashlib
            import sqlite3
            sqlite3.enable_callback_tracebacks(True)  # otherwise errors get ignored
            db_connection.connection.create_function("MD5", 1, lambda c: hashlib.md5(c.encode('utf8')).hexdigest())


class Lorem(FieldFunc):
    """
    Simple fixed-text scrubber, which replaces content with one paragraph of the well-known "lorem ipsum" text.
    """

    template = (
        "'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore "
        "et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
        "aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse "
        "cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in "
        "culpa qui officia deserunt mollit anim id est laborum.'")


class Concat(object):
    """
    Wrapper around django.db.functions.Concat for lazy concatenation of scrubbers.
    """

    def __init__(self, *expressions, **kwargs):
        self.expressions = expressions
        self.kwargs = kwargs

    def __call__(self, field):
        realized_expressions = list()
        for exp in self.expressions:
            realized_expressions.append(callable(exp) and exp(field) or exp)
        return DjangoConcat(*realized_expressions, **self.kwargs)


class Faker(object):
    INITIALIZED_PROVIDERS = set()

    def __init__(self, provider, *args, **kwargs):
        self.provider = provider
        self.provider_args = args
        self.provider_kwargs = kwargs
        args_hash = hash(self.provider_args) ^ hash(tuple(self.provider_kwargs.items()))
        self.provider_key = '%s - %s' % (self.provider, args_hash)

    def _initialize_data(self):
        from .models import FakeData

        # get locale from config and fall back to django's default one
        locale = settings_with_fallback('SCRUBBER_FAKER_LOCALE')
        if not locale:
            locale = to_locale(get_language())
        faker_instance = faker.Faker(locale=locale)

        # load additional faker providers
        for provider_name in settings_with_fallback('SCRUBBER_ADDITIONAL_FAKER_PROVIDERS'):
            # try to load module
            try:
                module_name, class_name = text(provider_name).rsplit('.', 1)
                module = importlib.import_module(module_name)
            except Exception:
                raise ScrubberInitError(
                    'module not found for provider defined in SCRUBBER_ADDITIONAL_FAKER_PROVIDERS: %s' % provider_name
                )

            # add provider to faker instance
            provider = getattr(module, class_name, None)
            if provider is None:
                raise ScrubberInitError(
                    'faker provider not found for provider defined in SCRUBBER_ADDITIONAL_FAKER_PROVIDERS: %s' %
                    provider_name)
            faker_instance.add_provider(provider)

        provider_args_str = ', '.join(str(i) for i in self.provider_args)
        provider_kwargs_str = ', '.join(str(i) for i in self.provider_kwargs)
        logger.info('Initializing fake scrub data for provider %s(%s, %s)',
                    self.provider, provider_args_str, provider_kwargs_str,
                    )
        # TODO: maybe be a bit smarter and only regenerate if needed?
        FakeData.objects.filter(provider=self.provider_key).delete()
        fakedata = []

        # if we don't reset the seed for each provider, registering a new one might change all
        # data for subsequent providers
        faker.Generator.seed(settings_with_fallback('SCRUBBER_RANDOM_SEED'))
        for i in range(settings_with_fallback('SCRUBBER_ENTRIES_PER_PROVIDER')):
            fakedata.append(FakeData(provider=self.provider_key, provider_offset=i,
                                     content=faker_instance.format(self.provider, *self.provider_args,
                                                                   **self.provider_kwargs)))

        try:
            FakeData.objects.bulk_create(fakedata)
        except IntegrityError as e:
            raise ScrubberInitError('Integrity error initializing faker data (%s); maybe decrease '
                                    'SCRUBBER_ENTRIES_PER_PROVIDER?' % (e,))

        self.INITIALIZED_PROVIDERS.add(self.provider_key)

    def __call__(self, field):
        """
        Lazily instantiate the actual subquery used for scrubbing.

        The Faker scrubber ignores the field parameter.
        """
        if self.provider_key not in self.INITIALIZED_PROVIDERS:
            self._initialize_data()

        # import it here to enable global scrubbers in settings.py
        from .models import FakeData

        return Cast(Subquery(FakeData.objects.filter(
            provider=self.provider_key,
            provider_offset=OuterRef('mod_pk')  # this outer field gets annotated before .update()
            # TODO: This can be used instead of the annotated mod_pk, as soon as this issue is fixed:
            # https://code.djangoproject.com/ticket/28621
            # This would allow us to have per-provider cardinality.
            # provider_offset=Mod(OuterRef('pk'), Subquery(FakeData.objects.provider_count(OuterRef('provider'))))
        ).values('content')[:1]), field)



import importlib
import logging
import warnings

from django.apps import apps
from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import FieldDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.db.models import F
from django.db.utils import IntegrityError, DataError

from ... import settings_with_fallback
from ...models import FakeData
from ...scrubbers import Keep
from ...services.validator import ScrubberValidatorService

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Replace database data according to model-specific or global scrubbing rules.'
    leave_locale_alone = True

    def add_arguments(self, parser):
        parser.add_argument('--model', type=str, required=False,
                            help='Scrub only a single model (format <app_label>.<model_name>)')
        parser.add_argument('--keep-sessions', action='store_true', required=False,
                            help='Will NOT truncate all (by definition critical) session data')
        parser.add_argument('--remove-fake-data', action='store_true', required=False,
                            help='Will truncate the database table storing preprocessed data for the Faker library. '
                                 'If you want to do multiple iterations of scrubbing, it will save you time to keep '
                                 'them. If not, you will add a huge bunch of data to your dump size.')

    def handle(self, *args, **kwargs):
        if not settings.DEBUG:
            # avoid logger, otherwise we might silently fail if we're on live and logging is being sent somewhere else
            self.stderr.write('This command should only be run with DEBUG=True, to avoid running on live systems')
            return False

        # Check STRICT mode
        if settings_with_fallback('SCRUBBER_STRICT_MODE'):
            service = ScrubberValidatorService()
            non_scrubbed_field_list = service.process()
            if len(non_scrubbed_field_list) > 0:
                self.stderr.write('When "SCRUBBER_STRICT_MODE" is enabled, you have to define a scrubbing policy '
                                  'for every text-based field.')
                return False

        global_scrubbers = settings_with_fallback('SCRUBBER_GLOBAL_SCRUBBERS')

        # run only for selected model
        if kwargs.get('model', None) is not None:
            app_label, model_name = kwargs.get('model').rsplit('.', 1)
            try:
                models = [apps.get_model(app_label=app_label, model_name=model_name)]
            except LookupError:
                raise CommandError('--model should be defined as <app_label>.<model_name>')

        # run for all models of all apps
        else:
            models = apps.get_models()

        scrubber_apps_list = settings_with_fallback('SCRUBBER_APPS_LIST')
        for model in models:
            if model._meta.proxy:
                continue
            if settings_with_fallback('SCRUBBER_SKIP_UNMANAGED') and not model._meta.managed:
                continue
            if scrubber_apps_list and model._meta.app_config.name not in scrubber_apps_list:
                continue

            scrubbers = dict()
            for field in model._meta.fields:
                if field.name in global_scrubbers:
                    scrubbers[field] = global_scrubbers[field.name]
                elif type(field) in global_scrubbers:
                    scrubbers[field] = global_scrubbers[type(field)]

            scrubbers.update(_get_model_scrubbers(model))

            # Filter out all fields marked as "to be kept"
            scrubbers_without_kept_fields = {}
            for field, scrubbing_method in scrubbers.items():
                if scrubbing_method != Keep:
                    scrubbers_without_kept_fields[field] = scrubbing_method
            scrubbers = scrubbers_without_kept_fields

            if not scrubbers:
                continue

            realized_scrubbers = _filter_out_disabled(_call_callables(scrubbers))

            logger.info('Scrubbing %s with %s', model._meta.label, realized_scrubbers)

            try:
                model.objects.annotate(
                    mod_pk=F('pk') % settings_with_fallback('SCRUBBER_ENTRIES_PER_PROVIDER')
                ).update(**realized_scrubbers)
            except IntegrityError as e:
                raise CommandError('Integrity error while scrubbing %s (%s); maybe increase '
                                   'SCRUBBER_ENTRIES_PER_PROVIDER?' % (model, e))
            except DataError as e:
                raise CommandError('DataError while scrubbing %s (%s)' % (model, e))

        # Truncate session data
        if not kwargs.get('keep_sessions', False):
            Session.objects.all().delete()

        # Truncate Faker data
        if kwargs.get('remove_fake_data', False):
            FakeData.objects.all().delete()


def _call_callables(d):
    """
    Helper to realize lazy scrubbers, like Faker, or global field-type scrubbers
    """
    return {k.name: (callable(v) and v(k) or v) for k, v in d.items()}


def _parse_scrubber_class_from_string(path: str):
    """
    Takes a string to a certain scrubber class and returns a python class definition - not an instance.
    """
    try:
        module_name, class_name = path.rsplit('.', 1)
        module = importlib.import_module(module_name)
        return getattr(module, class_name)
    except (ImportError, ValueError) as e:
        raise ImportError('Mapped scrubber class "%s" could not be found.' % path) from e


def _get_model_scrubbers(model):
    # Get model-scrubber-mapping from settings
    scrubber_mapping = settings_with_fallback('SCRUBBER_MAPPING')

    # Initialise scrubber list
    scrubbers = dict()

    # Check if model has a settings-defined...
    if model._meta.label in scrubber_mapping:
        scrubber_cls = _parse_scrubber_class_from_string(scrubber_mapping[model._meta.label])
    # If not...
    else:
        # Try to get the scrubber metaclass from the given model
        try:
            scrubber_cls = getattr(model, 'Scrubbers')
        except AttributeError:
            return scrubbers  # no model-specific scrubbers

    # Get field mappings from scrubber class
    for k, v in _get_fields(scrubber_cls):
        try:
            field = model._meta.get_field(k)
            scrubbers[field] = v
        except FieldDoesNotExist:
            warnings.warn(f'Scrubber defined for {model.__name__}.{k} but field does not exist')

    # Return scrubber-field-mapping
    return scrubbers


def _get_fields(d):
    """
    Helper to get "normal" (i.e.: non-magic and non-dunder) instance attributes.
    Returns an iterator of (field_name, field) tuples.
    """
    return ((k, v) for k, v in vars(d).items() if not k.startswith('_'))


def _filter_out_disabled(d):
    """
    Helper to remove Nones (actually any false-like type) from the scrubbers.
    This is needed so we can disable global scrubbers in a per-model basis.
    """
    return {k: v for k, v in d.items() if v}



from django.conf import settings
from django.db import models

__version__ = '1.2.1'

defaults = {
    'SCRUBBER_RANDOM_SEED': 42,  # we prefer idempotent scrubbing
    'SCRUBBER_ENTRIES_PER_PROVIDER': 1000,
    'SCRUBBER_GLOBAL_SCRUBBERS': {},
    'SCRUBBER_SKIP_UNMANAGED': True,
    'SCRUBBER_APPS_LIST': None,
    'SCRUBBER_ADDITIONAL_FAKER_PROVIDERS': [],
    'SCRUBBER_FAKER_LOCALE': None,
    'SCRUBBER_MAPPING': dict(),
    'SCRUBBER_STRICT_MODE': False,
    'SCRUBBER_REQUIRED_FIELD_TYPES': (
        models.CharField,
        models.TextField,
        models.URLField,
        models.JSONField,
        models.GenericIPAddressField,
        models.EmailField,
    ),
    'SCRUBBER_REQUIRED_FIELD_MODEL_WHITELIST': [
        'auth.Group',
        'auth.Permission',
        'contenttypes.ContentType',
        'sessions.Session',
        'sites.Site',
        'django_scrubber.FakeData',
    ],
}


# TODO: replace with ChainMap now that we only support py3
def settings_with_fallback(key):
    return getattr(settings, key, defaults[key])


class ScrubberInitError(Exception):
    pass
